/////////////////////////////////////////////////////////////////////
//Plugin for the PPionBeamAmplitude
//
//This plugin enables the amplitudes from Ref. 23 for the
//reaction pi + N -> N e+ e-
//
//The energy-dependent matrix elements have been supported by 
//the authors of Ref. 23 
//
//To use this plugin, type:
//makeDistributionManager()->Exec("pion_beam");
//A flat di-electron generator is automatially employed
//Therefore, the weights of the particles have to be used.
//
// 
//                             Author:  Schuldes/Froehlich
//                             Written: Jul 2009
//////////////////////////////////////////////////////////////////////

#include "PPionBeamPlugin.h"


PPionBeamPlugin::PPionBeamPlugin(const Char_t *id, const Char_t *de):
    PDistributionCollection(id, de) {

    Rho0Propagator  = NULL;
    OmegaPropagator = NULL;

    Pi_minusBeamAmplitude     = NULL;
    Pi_minusBeamAmplitude_gen = NULL;

    Pi_plusBeamAmplitude     = NULL;
    Pi_plusBeamAmplitude_gen = NULL;

    //    RequiresPlugin("elementary");
}

Bool_t PPionBeamPlugin::Activate(void) {

    PDistributionManagerUtil *pdmutil = makeDistributionManagerUtil();
    
    if (!makeStaticData()->IsParticleValid("pi- + p")) {
	makeStaticData()->AddParticle(14009, "pi- + p", 1.1);
    }
    if (!makeStaticData()->IsParticleValid("pi+ + n")) {
	makeStaticData()->AddParticle(13008, "pi+ + n", 1.1);
    }


    Int_t ipid[3];
    ipid[0] = makeStaticData()->GetParticleID("pi- + p");
    ipid[1] = makeStaticData()->GetParticleID("n");
    ipid[2] = makeStaticData()->GetParticleID("dilepton");

  
    //Add decay
    if (makeStaticData()->GetDecayKey(ipid, 2) < 0)
	makeStaticData()->AddDecay(-1, "pi- + p -> n + dilepton", 
				   "pi- + p","n,dilepton", 1.0 );
  
    ipid[0] = makeStaticData()->GetParticleID("pi+ + n");
    ipid[1] = makeStaticData()->GetParticleID("p");
    ipid[2] = makeStaticData()->GetParticleID("dilepton");


    if (makeStaticData()->GetDecayKey(ipid, 2) < 0)
	makeStaticData()->AddDecay(-1, "pi+ + n -> p + dilepton", 
				   "pi+ + n","p,dilepton", 1.0 );
   
    
    pdmutil->AddSubGroup("pion_beam", "Elementary pion beam reactions", "root");
    pdmutil->SetGroup("pion_beam");


    Pi_minusBeamAmplitude = 
	new PPionBeamAmplitude("Pi_minusBeamAmplitude@pi- + p_to_n_dilepton",
			       "Model from Lutz/Soyeur/Friman", -1);

    Pi_minusBeamAmplitude->Add("pi-,grandparent,pion");
    Pi_minusBeamAmplitude->Add("p,grandparent,p_in");
    Pi_minusBeamAmplitude->Add("dilepton,daughter");
    Pi_minusBeamAmplitude->Add("n,daughter,p_out");
    Pi_minusBeamAmplitude->Add("q,parent");

    PPionBeamAmplitude *Pi_plusBeamAmplitude = 
	new PPionBeamAmplitude("Pi_plusBeamAmplitude@pi+ + n_to_p_dilepton",
			       "Model from Lutz/Soyeur/Friman", -1);
  
    Pi_plusBeamAmplitude->Add("pi+,grandparent,pion");
    Pi_plusBeamAmplitude->Add("n,grandparent,p_in");
    Pi_plusBeamAmplitude->Add("dilepton,daughter");
    Pi_plusBeamAmplitude->Add("p,daughter,p_out");
    Pi_plusBeamAmplitude->Add("q,parent");

    TF1 *flat = new TF1("pion_flat", "1.", 0, 0.99);
    //The "PInclusiveModel" can be used as a generator:

    Pi_minusBeamAmplitude_gen = 
	new PInclusiveModel("pi-_flat@pi- + p_brems_n_dilepton/generator", "Dilepton generator", -1);
	
    //The distribution template:
    Pi_minusBeamAmplitude_gen->Add("q,parent");	
    Pi_minusBeamAmplitude_gen->Add("n,daughter");
    Pi_minusBeamAmplitude_gen->Add("dilepton,daughter,primary");
    Pi_minusBeamAmplitude_gen->SetSampleFunction(flat);
    //Enable distribution as a generator
    Pi_minusBeamAmplitude_gen->EnableGenerator();    
    pdmutil->Add(Pi_minusBeamAmplitude_gen);

    Pi_plusBeamAmplitude_gen = 
	new PInclusiveModel("pi+_flat@pi+ + n_brems_p_dilepton/generator", "Dilepton generator", -1);
	
    //The distribution template:
    Pi_plusBeamAmplitude_gen->Add("q,parent");	
    Pi_plusBeamAmplitude_gen->Add("p,daughter");
    Pi_plusBeamAmplitude_gen->Add("dilepton,daughter,primary");
    Pi_plusBeamAmplitude_gen->SetSampleFunction(flat);
    //Enable distribution as a generator
    Pi_plusBeamAmplitude_gen->EnableGenerator();    
    pdmutil->Add(Pi_plusBeamAmplitude_gen);

    
    double Graph_Rho_x[150] = {	1.200010,
				1.205343,
				1.210677,	
				1.216010,
				1.221343,
			 	1.226677,
			 	1.232010,	
			 	1.237343,
			 	1.242677,
			 	1.248010,
			 	1.253343,
				1.258677,
			 	1.264010,
			 	1.269343,
			 	1.274677,
			 	1.280010,
			 	1.285343,
			 	1.290677,
			 	1.296010,
			 	1.301343,
			 	1.306677,
			 	1.312010,
			 	1.317343,
			 	1.322677,
			 	1.328010,
			 	1.333343,
			 	1.338677,
			 	1.344010,
			 	1.349343,
			 	1.354677,
			 	1.360010,
			 	1.365343,
			 	1.370677,
			 	1.376010,
			 	1.381343,
			 	1.386677,
			 	1.392010,
			 	1.397343,
			 	1.402677,
			 	1.408010,
			 	1.413343,
			 	1.418677,
			 	1.424010,
			 	1.429343,
			 	1.434677,
			 	1.440010,
			 	1.445343,
			 	1.450677,
			 	1.456010,
			 	1.461343,
			 	1.466677,
			 	1.472010,
			 	1.477343,
			 	1.482677,
			 	1.488010,
			 	1.493343,
			 	1.498677,
			 	1.504010,
			 	1.509343,
			 	1.514677,
			 	1.520010,
			 	1.525343,
			 	1.530677,
			 	1.536010,
			 	1.541343,
			 	1.546677,
			 	1.552010,
			 	1.557343,
			 	1.562677,
			 	1.568010,
			 	1.573343,
			 	1.578677,
			 	1.584010,
			 	1.589343,
			 	1.594677,
			 	1.600010,
			 	1.605343,
			 	1.610677,
			 	1.616010,
			 	1.621343,
			 	1.626677,
			 	1.632010,
			 	1.637343,
			 	1.642677,
			 	1.648010,
			 	1.653343,
			 	1.658677,
			 	1.664010,
			 	1.669343,
			 	1.674677,
			 	1.680010,
			 	1.685343,
			 	1.690677,
			 	1.696010,
			 	1.701343,
			 	1.706677,
			 	1.712010,
			 	1.717343,
			 	1.722677,
			 	1.728010,
			 	1.733343,
			 	1.738677,
			 	1.744010,
			 	1.749343,
			 	1.754677,
			 	1.760010,
			 	1.765343,
			 	1.770677,
			 	1.776010,
			 	1.781343,
			 	1.786677,
			 	1.792010,
			 	1.797343,
			 	1.802677,
			 	1.808010,
			 	1.813343,
			 	1.818677,
			 	1.824010,
			 	1.829343,
			 	1.834677,
			 	1.840010,
			 	1.845343,
			 	1.850677,
			 	1.856010,
			 	1.861343,
			 	1.866677,
			 	1.872010,
			 	1.877343,
			 	1.882677,
			 	1.888010,
				1.893343,
			 	1.898677,
			 	1.904010,
			 	1.909343,
			 	1.914677,
			 	1.920010,
			 	1.925343,
			 	1.930677,
			 	1.936010,
			 	1.941343,
			 	1.946677,
			 	1.952010,
			 	1.957343,
			 	1.962677,
			 	1.968010,
			 	1.973343,
			 	1.978677,
			 	1.984010,
			 	1.989343,
			 	1.994677
    };
    double Graph_Rho_y_Re_T11[150] = {10.927144,
				      11.054252,
				      11.183319,
				      11.314396,
				      11.447535,
				      11.582795,
				      11.720238,
				      11.859932,
				      12.001948,
				      12.146364,
				      12.293266,
				      12.442745,
				      12.594900,
				      12.749843,
				      12.907689,
				      13.068568,
				      13.232624,
				      13.400007,
				      13.570890,
				      13.745453,
				      13.923903,
				      14.106458,
				      14.293363,
				      14.484882,
				      14.681305,
				      14.882954,
				      15.090174,
				      15.303348,
				      15.522894,
				      15.749271,
				      15.982983,
				      16.224578,
				      16.474674,
				      16.733924,
				      17.003239,
				      17.283150,
				      17.574746,
				      17.879095,
				      18.197160,
				      18.530794,
				      18.881362,
				      19.250767,
				      19.641255,
				      20.055519,
				      20.496921,
				      20.969657,
				      21.479111,
				      22.032411,
				      22.639301,
				      23.313768,
				      24.077115,
				      24.964850,
				      26.046298,
				      27.514324,
				      23.429189,
				      17.922962,
				      14.284720,
				      11.418496,
				      9.060166,
				      7.095792,
				      5.459748,
				      4.106839,
				      3.001836,
				      2.115744,
				      1.424028,
				      0.905675,
				      0.542946,
				      0.320928,
				      0.227112,
				      0.251003,
				      0.383341,
				      0.615095,
				      0.936011,
				      1.331964,
				      1.780537,
				      2.241361,
				      2.625383,
				      2.419523,
				      2.773354,
				      2.867063,
				      2.729286,
				      2.428496,
				      2.038813,
				      1.616750,
				      1.198547,
				      0.805547,
				      0.449584,
				      0.137842,
				      -0.124826,
				      -0.334509,
				      -0.488020,
				      -0.569560,
				      -0.491498,
				      -0.384744,
				      -0.250865,
				      -0.088530,
				      0.109346,
				      0.357700,
				      0.607211,
				      0.710514,
				      0.734399,
				      0.729459,
				      0.711516,
				      0.686936,
				      0.659246,
				      0.629241,
				      0.597348,
				      0.565423,
				      0.534413,
				      0.504613,
				      0.476116,
				      0.449040,
				      0.423345,
				      0.399003,
				      0.376021,
				      0.354327,
				      0.333904,
				      0.314705,
				      0.296628,
				      0.279665,
				      0.263739,
				      0.248807,
				      0.234862,
				      0.221789,
				      0.209578,
				      0.198175,
				      0.187539,
				      0.177665,
				      0.168470,
				      0.159933,
				      0.152025,
				      0.144707,
				      0.137973,
				      0.131771,
				      0.126073,
				      0.120865,
				      0.116114,
				      0.111810,
				      0.107930,
				      0.104436,
				      0.101328,
				      0.098576,
				      0.096168,
				      0.094107,
				      0.092346,
				      0.090888,
				      0.089713,
				      0.088812,
				      0.088187,
				      0.087804 };

    double Graph_Rho_y_Im_T11[150] = {0.910200,
				      0.954286,
				      0.999784,
				      1.046767,
				      1.095314,
				      1.145507,
				      1.197431,
				      1.251175,
				      1.306835,
				      1.364512,
				      1.424314,
				      1.486355,
				      1.550759,
				      1.617661,
				      1.687204,
				      1.759545,
				      1.834858,
				      1.913330,
				      1.995166,
				      2.080593,
				      2.169870,
				      2.263268,
				      2.361107,
				      2.463729,
				      2.571525,
				      2.684937,
				      2.804456,
				      2.930638,
				      3.064111,
				      3.205591,
				      3.355889,
				      3.515921,
				      3.686747,
				      3.869556,
				      4.065889,
				      4.277049,
				      4.505037,
				      4.752027,
				      5.020286,
				      5.313351,
				      5.634549,
				      5.988255,
				      6.379784,
				      6.815732,
				      7.304491,
				      7.856877,
				      8.487220,
				      9.215156,
				      10.068662,
				      11.089791,
				      12.346527,
				      13.960898,
				      16.194830,
				      19.872674,
				      28.166761,
				      28.334894,
				      27.899177,
				      27.148490,
				      26.189925,
				      25.090924,
				      23.901309,
				      22.659430,
				      21.395293,
				      20.132271,
				      18.888536,
				      17.678337,
				      16.513072,
				      15.402340,
				      14.355056,
				      13.380355,
				      12.488697,
				      11.692907,
				      11.009356,
				      10.459301,
				      10.070405,
				      9.879183,
				      9.937317,
				      10.098127,
				      9.923171,
				      10.131774,
				      10.360310,
				      10.505713,
				      10.533933,
				      10.447338,
				      10.262323,
				      9.997361,
				      9.669314,
				      9.290949,
				      8.873061,
				      8.423727,
				      7.948418,
				      7.407230,
				      6.941536,
				      6.545564,
				      6.197589,
				      5.892170,
				      5.632294,
				      5.444328,
				      5.414461,
				      5.460242,
				      5.471005,
				      5.452900,
				      5.417032,
				      5.370151,
				      5.315913,
				      5.258264,
				      5.196802,
				      5.132822,
				      5.067713,
				      5.002415,
				      4.937646,
				      4.873792,
				      4.811268,
				      4.750327,
				      4.691095,
				      4.633722,
				      4.578238,
				      4.524679,
				      4.473112,
				      4.423478,
				      4.375775,
				      4.329958,
				      4.285930,
				      4.243713,
				      4.203203,
				      4.164347,
				      4.127078,
				      4.091288,
				      4.056955,
				      4.023991,
				      3.992322,
				      3.961887,
				      3.932592,
				      3.904393,
				      3.877226,
				      3.851016,
				      3.825713,
				      3.801244,
				      3.777555,
				      3.754608,
				      3.732333,
				      3.710694,
				      3.689639,
				      3.669109,
				      3.649084,
				      3.629514,
				      3.610366,
				      3.591602,
				      3.573178,
				      3.555080			 
    };
  
    double Graph_Rho_y_Re_T13[150] = {79.999754,
				      81.202991,
				      82.440358,
				      83.713773,
				      85.025272,
				      86.377041,
				      87.771493,
				      89.211189,
				      90.698927,
				      92.237830,
				      93.831169,
				      95.482617,
				      97.196182,
				      98.976049,
				      100.827450,
				      102.755150,
				      104.765348,
				      106.864575,
				      109.059305,
				      111.358697,
				      113.770248,
				      116.305084,
				      118.973334,
				      121.787054,
				      124.760419,
				      127.907485,
				      131.245170,
				      134.791346,
				      138.565223,
				      142.588566,
				      146.884262,
				      151.478873,
				      156.398289,
				      161.676925,
				      167.341747,
				      173.434967,
				      179.996182,
				      187.063240,
				      194.688249,
				      202.906814,
				      211.768987,
				      221.314068,
				      231.557608,
				      242.501176,
				      254.099687,
				      266.228576,
				      278.660484,
				      290.947852,
				      302.405098,
				      311.914614,
				      317.814662,
				      317.820817,
				      309.052088,
				      288.515445,
				      254.048605,
				      205.724301,
				      146.838922,
				      83.613864,
				      23.084481,
				      -29.164873,
				      -70.385655,
				      -100.443169,
				      -120.732715,
				      -133.264735,
				      -140.019930,
				      -142.638508,
				      -142.414483,
				      -140.284675,
				      -136.924826,
				      -132.806143,
				      -128.252667,
				      -123.487439,
				      -118.658135,
				      -113.863949,
				      -109.167613,
				      -104.608551,
				      -100.210174,
				      -95.983230,
				      -91.931735,
				      -88.053837,
				      -84.344307,
				      -80.796371,
				      -77.400747,
				      -74.148347,
				      -71.027939,
				      -68.029386,
				      -65.142647,
				      -62.353245,
				      -59.652588,
				      -57.027252,
				      -54.463349,
				      -51.945514,
				      -49.455479,
				      -46.968581,
				      -44.449700,
				      -41.840615,
				      -39.030936,
				      -35.786804,
				      -32.178549,
				      -29.528051,
				      -27.653129,
				      -26.154901,
				      -24.882372,
				      -23.767237,
				      -22.768257,
				      -21.868339,
				      -21.064489,
				      -20.333940,
				      -19.661180,
				      -19.037930,
				      -18.458681,
				      -17.918412,
				      -17.413810,
				      -16.941920,
				      -16.499816,
				      -16.085387,
				      -15.696490,
				      -15.331194,
				      -14.988424,
				      -14.666358,
				      -14.363838,
				      -14.079714,
				      -13.812463,
				      -13.561723,
				      -13.326213,
				      -13.105173,
				      -12.897808,
				      -12.703081,
				      -12.520644,
				      -12.349726,
				      -12.189612,
				      -12.039848,
				      -11.899627,
				      -11.768584,
				      -11.646277,
				      -11.532149,
				      -11.426017,
				      -11.327059,
				      -11.234826,
				      -11.148986,
				      -11.069003,
				      -10.994557,
				      -10.925266,
				      -10.860516,
				      -10.800336,
				      -10.744141,
				      -10.691750,
				      -10.642770,
				      -10.596803,
				      -10.553743
    };

    double Graph_Rho_y_Im_T13[150] = {0.172797,
				      0.202131,
				      0.235426,
				      0.273128,
				      0.315725,
				      0.363733,
				      0.417710,
				      0.478276,
				      0.546117,
				      0.622003,
				      0.706793,
				      0.801454,
				      0.907079,
				      1.024912,
				      1.156358,
				      1.303057,
				      1.466877,
				      1.649993,
				      1.855002,
				      2.084838,
				      2.343180,
				      2.634180,
				      2.963042,
				      3.335988,
				      3.760336,
				      4.245326,
				      4.801695,
				      5.442363,
				      6.183616,
				      7.043319,
				      8.043708,
				      9.212110,
				      10.577915,
				      12.181168,
				      14.062523,
				      16.279532,
				      18.897777,
				      21.995371,
				      25.675367,
				      30.055224,
				      35.290852,
				      41.573748,
				      49.136717,
				      58.278429,
				      69.369187,
				      82.861222,
				      99.319087,
				      119.371105,
				      143.764923,
				      173.220093,
				      208.276747,
				      249.049908,
				      294.575917,
				      342.478895,
				      388.404311,
				      426.509539,
				      451.047793,
				      458.502875,
				      449.102019,
				      426.338072,
				      395.230875,
				      360.487558,
				      325.592027,
				      292.650111,
				      262.673758,
				      236.014050,
				      212.576840,
				      192.110301,
				      174.272132,
				      158.713074,
				      145.122917,
				      133.200461,
				      122.713454,
				      113.442718,
				      105.210160,
				      97.871589,
				      91.297458,
				      85.382517,
				      80.038571,
				      75.188341,
				      70.772272,
				      66.731762,
				      63.021465,
				      59.603185,
				      56.439346,
				      53.505943,
				      50.773808,
				      48.217359,
				      45.822452,
				      43.568704,
				      41.442980,
				      39.429563,
				      37.515746,
				      35.691573,
				      33.946246,
				      32.276516,
				      30.702591,
				      29.384301,
				      29.135239,
				      29.803776,
				      30.387191,
				      30.800782,
				      31.084135,
				      31.273437,
				      31.394989,
				      31.471681,
				      31.500655,
				      31.483502,
				      31.429924,
				      31.347404,
				      31.241453,
				      31.116455,
				      30.975392,
				      30.820882,
				      30.655076,
				      30.479455,
				      30.295663,
				      30.104566,
				      29.907259,
				      29.704468,
				      29.496734,
				      29.284797,
				      29.068989,
				      28.849764,
				      28.627389,
				      28.402126,
				      28.174274,
				      27.944018,
				      27.711490,
				      27.476881,
				      27.240225,
				      27.001759,
				      26.761511,
				      26.519548,
				      26.276010,
				      26.031002,
				      25.784291,
				      25.535902,
				      25.285895,
				      25.034338,
				      24.781301,
				      24.526823,
				      24.270938,
				      24.013808,
				      23.755328,
				      23.495667,
				      23.234796,
				      22.972822,
				      22.709809,
				      22.445771
    };

    double Graph_Rho_y_Re_T31[150] = {-18.898530,
				      -18.954238,
				      -19.008932,
				      -19.062622,
				      -19.115336,
				      -19.167093,
				      -19.217916,
				      -19.267833,
				      -19.316876,
				      -19.365079,
				      -19.412470,
				      -19.459096,
				      -19.504998,
				      -19.550217,
				      -19.594806,
				      -19.638821,
				      -19.682309,
				      -19.725339,
				      -19.767980,
				      -19.810299,
				      -19.852371,
				      -19.894290,
				      -19.936132,
				      -19.978009,
				      -20.020017,
				      -20.062275,
				      -20.104905,
				      -20.148048,
				      -20.191848,
				      -20.236468,
				      -20.282075,
				      -20.328867,
				      -20.377060,
				      -20.426869,
				      -20.478591,
				      -20.532417,
				      -20.588689,
				      -20.647744,
				      -20.709872,
				      -20.775615,
				      -20.845352,
				      -20.919569,
				      -20.998812,
				      -21.083699,
				      -21.174885,
				      -21.273119,
				      -21.379240,
				      -21.494175,
				      -21.618967,
				      -21.754770,
				      -21.902869,
				      -22.064711,
				      -22.241895,
				      -22.436216,
				      -22.649648,
				      -22.884333,
				      -23.142670,
				      -23.427171,
				      -23.740492,
				      -24.085340,
				      -24.464253,
				      -24.879401,
				      -25.332139,
				      -25.822282,
				      -26.347212,
				      -26.900283,
				      -27.468808,
				      -28.031350,
				      -28.554390,
				      -28.989253,
				      -29.270353,
				      -29.317394,
				      -29.044877,
				      -28.380249,
				      -27.288896,
				      -25.795851,
				      -23.990377,
				      -22.009025,
				      -19.999765,
				      -18.088432,
				      -16.359488,
				      -14.852245,
				      -13.572492,
				      -12.503751,
				      -11.618787,
				      -10.887591,
				      -10.280414,
				      -9.771085,
				      -9.336666,
				      -8.955685,
				      -8.602002,
				      -8.373614,
				      -8.381858,
				      -8.305447,
				      -8.198825,
				      -8.075954,
				      -7.943039,
				      -7.803117,
				      -7.658536,
				      -7.510620,
				      -7.360422,
				      -7.208926,
				      -7.056351,
				      -6.903115,
				      -6.750076,
				      -6.596933,
				      -6.444168,
				      -6.291804,
				      -6.140002,
				      -5.988871,
				      -5.838466,
				      -5.688884,
				      -5.540139,
				      -5.392265,
				      -5.245312,
				      -5.099299,
				      -4.954260,
				      -4.810212,
				      -4.667167,
				      -4.525154,
				      -4.384183,
				      -4.244278,
				      -4.105449,
				      -3.967714,
				      -3.831086,
				      -3.695582,
				      -3.561217,
				      -3.427992,
				      -3.295935,
				      -3.165056,
				      -3.035363,
				      -2.906875,
				      -2.779587,
				      -2.653525,
				      -2.528700,
				      -2.405113,
				      -2.282784,
				      -2.161710,
				      -2.041901,
				      -1.923378,
				      -1.806136,
				      -1.690193,
				      -1.575549,
				      -1.462195,
				      -1.350166,
				      -1.239448,
				      -1.130057,
				      -1.021995,
				      -0.915245,
				      -0.809840
    };


    double Graph_Rho_y_Im_T31[150] = {5.177131,
				      5.337784,
				      5.497719,
				      5.656976,
				      5.815590,
				      5.973591,
				      6.131000,
				      6.287828,
				      6.444084,
				      6.599785,
				      6.754926,
				      6.909510,
				      7.063533,
				      7.216990,
				      7.369875,
				      7.522178,
				      7.673890,
				      7.824999,
				      7.975490,
				      8.125350,
				      8.274567,
				      8.423120,
				      8.570994,
				      8.718172,
				      8.864631,
				      9.010357,
				      9.155329,
				      9.299518,
				      9.442903,
				      9.585461,
				      9.727158,
				      9.867959,
				      10.007833,
				      10.146736,
				      10.284625,
				      10.421424,
				      10.557077,
				      10.691512,
				      10.824623,
				      10.956336,
				      11.086503,
				      11.214980,
				      11.341581,
				      11.466103,
				      11.588274,
				      11.707783,
				      11.824250,
				      11.937219,
				      12.046145,
				      12.150328,
				      12.248938,
				      12.340956,
				      12.425123,
				      12.499879,
				      12.563336,
				      12.613054,
				      12.646085,
				      12.658707,
				      12.646276,
				      12.603044,
				      12.521635,
				      12.392996,
				      12.205773,
				      11.945939,
				      11.596455,
				      11.136726,
				      10.543048,
				      9.789461,
				      8.850032,
				      7.703896,
				      6.342526,
				      4.780726,
				      3.068518,
				      1.297886,
				      -0.401210,
				      -1.882716,
				      -3.019820,
				      -3.736728,
				      -4.024632,
				      -3.932383,
				      -3.542528,
				      -2.946559,
				      -2.226248,
				      -1.446060,
				      -0.651834,
				      0.126823,
				      0.872474,
				      1.577026,
				      2.239358,
				      2.864945,
				      3.474202,
				      4.270884,
				      4.721048,
				      5.079860,
				      5.392000,
				      5.671536,
				      5.925894,
				      6.159182,
				      6.374945,
				      6.575360,
				      6.762250,
				      6.937481,
				      7.101472,
				      7.255024,
				      7.400154,
				      7.536142,
				      7.664262,
				      7.784613,
				      7.897715,
				      8.004005,
				      8.103720,
				      8.197358,
				      8.285037,
				      8.366975,
				      8.443519,
				      8.514807,
				      8.581137,
				      8.642709,
				      8.699571,
				      8.752011,
				      8.800131,
				      8.844110,
				      8.884206,
				      8.920372,
				      8.952839,
				      8.981702,
				      9.007089,
				      9.029199,
				      9.048028,
				      9.063717,
				      9.076381,
				      9.086097,
				      9.093019,
				      9.097181,
				      9.098662,
				      9.097581,
				      9.093991,
				      9.088002,
				      9.079685,
				      9.069073,
				      9.056272,
				      9.041333,
				      9.024324,
				      9.005355,
				      8.984416,
				      8.961603,
				      8.936963,
				      8.910553,
				      8.882464,
				      8.852699
    };
  
    double Graph_Rho_y_Re_T33[150] = {-44.165564,
				      -44.782299,
				      -45.413780,
				      -46.060859,
				      -46.724507,
				      -47.405805,
				      -48.105886,
				      -48.826045,
				      -49.567696,
				      -50.332251,
				      -51.121502,
				      -51.937242,
				      -52.781461,
				      -53.656910,
				      -54.564928,
				      -55.509720,
				      -56.493338,
				      -57.518695,
				      -58.592089,
				      -59.712470,
				      -60.889741,
				      -62.121776,
				      -63.416569,
				      -64.779063,
				      -66.206418,
				      -67.708315,
				      -69.279943,
				      -70.916366,
				      -72.631021,
				      -74.399175,
				      -76.218163,
				      -78.099669,
				      -79.999924,
				      -81.952793,
				      -83.886394,
				      -85.837848,
				      -87.787357,
				      -89.698700,
				      -91.590760,
				      -93.410871,
				      -95.168676,
				      -96.850432,
				      -98.419853,
				      -99.869159,
				      -101.181770,
				      -102.333399,
				      -103.319808,
				      -104.099973,
				      -104.679432,
				      -105.037907,
				      -105.159607,
				      -105.044851,
				      -104.676379,
				      -104.060251,
				      -103.194744,
				      -102.084690,
				      -100.737332,
				      -99.165891,
				      -97.380714,
				      -95.400311,
				      -93.244871,
				      -90.927100,
				      -88.471985,
				      -85.899878,
				      -83.220662,
				      -80.468165,
				      -77.645458,
				      -74.776600,
				      -71.872527,
				      -68.940500,
				      -66.002479,
				      -63.049322,
				      -60.105192,
				      -57.162796,
				      -54.223449,
				      -51.296106,
				      -48.375228,
				      -45.459236,
				      -42.548048,
				      -39.633488,
				      -36.720837,
				      -33.796985,
				      -30.858233,
				      -27.903517,
				      -24.922299,
				      -21.926087,
				      -18.902438,
				      -15.856397,
				      -12.793796,
				      -9.724366,
				      -6.661738,
				      -3.613763,
				      -0.608983,
				      2.355058,
				      5.210696,
				      7.963255,
				      10.582684,
				      13.018834,
				      15.262771,
				      17.240957,
				      18.961181,
				      20.385818,
				      21.449639,
				      22.202664,
				      22.600870,
				      22.667623,
				      22.485054,
				      21.999008,
				      21.328073,
				      20.501632,
				      19.547779,
				      18.532754,
				      17.446044,
				      16.354439,
				      15.252220,
				      14.159107,
				      13.093216,
				      12.046872,
				      11.035352,
				      10.057886,
				      9.112631,
				      8.205265,
				      7.331877,
				      6.491906,
				      5.687883,
				      4.912460,
				      4.171532,
				      3.456577,
				      2.770329,
				      2.112600,
				      1.478096,
				      0.869541,
				      0.282205,
				      -0.282582,
				      -0.826144,
				      -1.350564,
				      -1.855644,
				      -2.342791,
				      -2.813130,
				      -3.265676,
				      -3.704098,
				      -4.126980,
				      -4.534898,
				      -4.929698,
				      -5.310221,
				      -5.679083,
				      -6.034973,
				      -6.378910,
				      -6.711708,
				      -7.032907
    };
  
  
    double Graph_Rho_y_Im_T33[150] = {-0.009574,
				      -0.011236,
				      -0.013128,
				      -0.015280,
				      -0.017727,
				      -0.020566,
				      -0.023942,
				      -0.028041,
				      -0.033098,
				      -0.039414,
				      -0.047368,
				      -0.057439,
				      -0.070236,
				      -0.086530,
				      -0.107291,
				      -0.133759,
				      -0.167492,
				      -0.210456,
				      -0.265179,
				      -0.334718,
				      -0.423091,
				      -0.535009,
				      -0.676495,
				      -0.854774,
				      -1.078020,
				      -1.356318,
				      -1.700164,
				      -2.120676,
				      -2.631449,
				      -3.242145,
				      -3.964043,
				      -4.811083,
				      -5.785905,
				      -6.906905,
				      -8.165078,
				      -9.579957,
				      -11.155927,
				      -12.887888,
				      -14.794431,
				      -16.858743,
				      -19.095246,
				      -21.507339,
				      -24.079181,
				      -26.813884,
				      -29.707263,
				      -32.743940,
				      -35.930967,
				      -39.219763,
				      -42.625779,
				      -46.121508,
				      -49.674060,
				      -53.290255,
				      -56.901223,
				      -60.518525,
				      -64.107220,
				      -67.635337,
				      -71.090745,
				      -74.436419,
				      -77.669200,
				      -80.760091,
				      -83.692600,
				      -86.467419,
				      -89.065923,
				      -91.483225,
				      -93.726513,
				      -95.780370,
				      -97.660409,
				      -99.363551,
				      -100.898416,
				      -102.273414,
				      -103.491548,
				      -104.567807,
				      -105.504611,
				      -106.314090,
				      -107.003055,
				      -107.577203,
				      -108.045087,
				      -108.410391,
				      -108.678139,
				      -108.849840,
				      -108.926569,
				      -108.909725,
				      -108.794662,
				      -108.579441,
				      -108.252061,
				      -107.808841,
				      -107.235865,
				      -106.519939,
				      -105.651462,
				      -104.611737,
				      -103.391656,
				      -101.976705,
				      -100.352962,
				      -98.518143,
				      -96.457731,
				      -94.174990,
				      -91.670379,
				      -88.946940,
				      -86.017365,
				      -82.906955,
				      -79.632547,
				      -76.233013,
				      -72.772859,
				      -69.273632,
				      -65.821844,
				      -62.468224,
				      -59.215597,
				      -56.189297,
				      -53.345014,
				      -50.709890,
				      -48.304668,
				      -46.084162,
				      -44.094868,
				      -42.259156,
				      -40.597342,
				      -39.083314,
				      -37.686660,
				      -36.417188,
				      -35.242315,
				      -34.155783,
				      -33.152847,
				      -32.214421,
				      -31.338977,
				      -30.519467,
				      -29.742801,
				      -29.015464,
				      -28.319097,
				      -27.663181,
				      -27.036550,
				      -26.433956,
				      -25.860077,
				      -25.305275,
				      -24.773761,
				      -24.259292,
				      -23.760774,
				      -23.278570,
				      -22.809698,
				      -22.353615,
				      -21.909736,
				      -21.474249,
				      -21.051088,
				      -20.636091,
				      -20.228549,
				      -19.829581,
				      -19.436313,
				      -19.051205,
				      -18.671257,
				      -18.296820,
				      -17.928025,
				      -17.563446
    };

    double Graph_Om_x[150] = { 1.200010,
			       1.205343,
			       1.210677,
			       1.216010,
			       1.221343,
			       1.226677,
			       1.232010,
			       1.237343,
			       1.242677,
			       1.248010,
			       1.253343,
			       1.258677,
			       1.264010,
			       1.269343,
			       1.274677,
			       1.280010,
			       1.285343,
			       1.290677,
			       1.296010,
			       1.301343,
			       1.306677,
			       1.312010,
			       1.317343,
			       1.322677,
			       1.328010,
			       1.333343,
			       1.338677,
			       1.344010,
			       1.349343,
			       1.354677,
			       1.360010,
			       1.365343,
			       1.370677,
			       1.376010,
			       1.381343,
			       1.386677,
			       1.392010,
			       1.397343,
			       1.402677,
			       1.408010,
			       1.413343,
			       1.418677,
			       1.424010,
			       1.429343,
			       1.434677,
			       1.440010,
			       1.445343,
			       1.450677,
			       1.456010,
			       1.461343,
			       1.466677,
			       1.472010,
			       1.477343,
			       1.482677,
			       1.488010,
			       1.493343,
			       1.498677,
			       1.504010,
			       1.509343,
			       1.514677,
			       1.520010,
			       1.525343,
			       1.530677,
			       1.536010,
			       1.541343,
			       1.546677,
			       1.552010,
			       1.557343,
			       1.562677,
			       1.568010,
			       1.573343,
			       1.578677,
			       1.584010,
			       1.589343,
			       1.594677,
			       1.600010,
			       1.605343,
			       1.610677,
			       1.616010,
			       1.621343,
			       1.626677,
			       1.632010,
			       1.637343,
			       1.642677,
			       1.648010,
			       1.653343,
			       1.658677,
			       1.664010,
			       1.669343,
			       1.674677,
			       1.680010,
			       1.685343,
			       1.690677,
			       1.696010,
			       1.701343,
			       1.706677,
			       1.712010,
			       1.717343,
			       1.722677,
			       1.728010,
			       1.733343,
			       1.738677,
			       1.744010,
			       1.749343,
			       1.754677,
			       1.760010,
			       1.765343,
			       1.770677,
			       1.776010,
			       1.781343,
			       1.786677,
			       1.792010,
			       1.797343,
			       1.802677,
			       1.808010,
			       1.813343,
			       1.818677,
			       1.824010,
			       1.829343,
			       1.834677,
			       1.840010,
			       1.845343,
			       1.850677,
			       1.856010,
			       1.861343,
			       1.866677,
			       1.872010,
			       1.877343,
			       1.882677,
			       1.888010,
			       1.893343,
			       1.898677,
			       1.904010,
			       1.909343,
			       1.914677,
			       1.920010,
			       1.925343,
			       1.930677,
			       1.936010,
			       1.941343,
			       1.946677,
			       1.952010,
			       1.957343,
			       1.962677,
			       1.968010,
			       1.973343,
			       1.978677,
			       1.984010,
			       1.989343,
			       1.994677
    };
	




	   
  
    double Graph_Om_y_Im_T11 [150] = {   1.389068,
					 1.456658,
					 1.526448,
					 1.598560,
					 1.673117,
					 1.750252,
					 1.830102,
					 1.912813,
					 1.998538,
					 2.087443,
					 2.179702,
					 2.275503,
					 2.375050,
					 2.478561,
					 2.586272,
					 2.698442,
					 2.815355,
					 2.937317,
					 3.064668,
					 3.197779,
					 3.337072,
					 3.482994,
					 3.636067,
					 3.796850,
					 3.965981,
					 4.144179,
					 4.332243,
					 4.531073,
					 4.741692,
					 4.965255,
					 5.203077,
					 5.456638,
					 5.727655,
					 6.018059,
					 6.330312,
					 6.666592,
					 7.030124,
					 7.424451,
					 7.853343,
					 8.322490,
					 8.837434,
					 9.405405,
					 10.035236,
					 10.737938,
					 11.527571,
					 12.422345,
					 13.446496,
					 14.633362,
					 16.030622,
					 17.710142,
					 19.788316,
					 22.474312,
					 26.216790,
					 32.424445,
					 46.772593,
					 47.439029,
					 47.041099,
					 46.096837,
					 44.789221,
					 43.229954,
					 41.499436,
					 39.658386,
					 37.752824,
					 35.816881,
					 33.874883,
					 31.942785,
					 30.029712,
					 28.138978,
					 26.268922,
					 24.413804,
					 22.563957,
					 20.706500,
					 18.825883,
					 16.904263,
					 14.923372,
					 12.868877,
					 10.750605,
					 10.238991,
					 10.752178,
					 10.486870,
					 10.780760,
					 11.675000,
					 13.002159,
					 14.575257,
					 16.250461,
					 17.933620,
					 19.568770,
					 21.127969,
					 22.606847,
					 24.033553,
					 25.541496,
					 27.358317,
					 26.236157,
					 25.026124,
					 23.668561,
					 22.153756,
					 20.451159,
					 18.491888,
					 16.540051,
					 15.356865,
					 14.610553,
					 14.046350,
					 13.581249,
					 13.180783,
					 12.827425,
					 12.512975,
					 12.234395,
					 11.979383,
					 11.741607,
					 11.518226,
					 11.307183,
					 11.106938,
					 10.916210,
					 10.733968,
					 10.559365,
					 10.391660,
					 10.230250,
					 10.074594,
					 9.924207,
					 9.778691,
					 9.637670,
					 9.500835,
					 9.367901,
					 9.238590,
					 9.112678,
					 8.989949,
					 8.870220,
					 8.753314,
					 8.639065,
					 8.527328,
					 8.417973,
					 8.310875,
					 8.205922,
					 8.103007,
					 8.002030,
					 7.902903,
					 7.805546,
					 7.709879,
					 7.615831,
					 7.523333,
					 7.432321,
					 7.342739,
					 7.254531,
					 7.167650,
					 7.082044,
					 6.997669,
					 6.914482,
					 6.832444,
					 6.751522,
					 6.671677
    };


    double Graph_Om_y_Re_T11 [150] = {  16.676058,
					16.873620,
					17.074454,
					17.278667,
					17.486368,
					17.697678,
					17.912727,
					18.131652,
					18.354605,
					18.581743,
					18.813243,
					19.049291,
					19.290092,
					19.535868,
					19.786854,
					20.043311,
					20.305525,
					20.573799,
					20.848469,
					21.129896,
					21.418482,
					21.714654,
					22.018892,
					22.331705,
					22.653656,
					22.985366,
					23.327505,
					23.680812,
					24.046094,
					24.424238,
					24.816224,
					25.223121,
					25.646128,
					26.086550,
					26.546026,
					27.025899,
					27.528193,
					28.055032,
					28.608615,
					29.192193,
					29.808795,
					30.462277,
					31.157229,
					31.899187,
					32.695000,
					33.553233,
					34.484872,
					35.504376,
					36.631442,
					37.894109,
					39.334727,
					41.023316,
					43.095027,
					45.921357,
					39.521110,
					30.351765,
					24.284479,
					19.471935,
					15.470932,
					12.091298,
					9.223901,
					6.793776,
					4.742290,
					3.020930,
					1.588384,
					0.409093,
					-0.546716,
					-1.303340,
					-1.879366,
					-2.286958,
					-2.531089,
					-2.607636,
					-2.500004,
					-2.173386,
					-1.561956,
					-0.536450,
					1.219817,
					5.427883,
					5.767103,
					7.351354,
					9.212795,
					10.915352,
					12.236154,
					13.106596,
					13.540076,
					13.580716,
					13.277968,
					12.671440,
					11.791697,
					10.647630,
					9.199456,
					5.813249,
					2.983655,
					0.971801,
					-0.634399,
					-1.927270,
					-2.916273,
					-3.479861,
					-3.186752,
					-2.468163,
					-1.900464,
					-1.464063,
					-1.112943,
					-0.819212,
					-0.565878,
					-0.339191,
					-0.142748,
					0.027742,
					0.178540,
					0.313819,
					0.436449,
					0.548524,
					0.651670,
					0.747157,
					0.835989,
					0.918994,
					0.996849,
					1.070123,
					1.139306,
					1.204797,
					1.266949,
					1.326061,
					1.382391,
					1.436179,
					1.487623,
					1.536891,
					1.584151,
					1.629527,
					1.673151,
					1.715135,
					1.755573,
					1.794553,
					1.832154,
					1.868451,
					1.903506,
					1.937379,
					1.970120,
					2.001780,
					2.032400,
					2.062027,
					2.090689,
					2.118424,
					2.145266,
					2.171235,
					2.196363,
					2.220677,
					2.244195,
					2.266939,
					2.288926,
					2.310179
    };
  
    double Graph_Om_y_Re_T13 [150] = { 145.316208,
				       148.083055,
				       150.944508,
				       153.906334,
				       156.974935,
				       160.157338,
				       163.460977,
				       166.894114,
				       170.465700,
				       174.184960,
				       178.062835,
				       182.110576,
				       186.340514,
				       190.768185,
				       195.404279,
				       200.270501,
				       205.381185,
				       210.754440,
				       216.420894,
				       222.386461,
				       228.695727,
				       235.353877,
				       242.401765,
				       249.871843,
				       257.775024,
				       266.165408,
				       275.056849,
				       284.469814,
				       294.485659,
				       305.082245,
				       316.311351,
				       328.273510,
				       340.923025,
				       354.441140,
				       368.722597,
				       383.974977,
				       400.255158,
				       417.575487,
				       436.113490,
				       455.826814,
				       476.852325,
				       499.234042,
				       522.883651,
				       547.731147,
				       573.543413,
				       599.834811,
				       625.891083,
				       650.281582,
				       671.111683,
				       685.303913,
				       688.469236,
				       674.899026,
				       637.580290,
				       569.892821,
				       467.874905,
				       333.701202,
				       177.649521,
				       16.540020,
				       -132.003962,
				       -255.160247,
				       -347.740621,
				       -411.019480,
				       -449.700058,
				       -469.555335,
				       -475.915437,
				       -473.013093,
				       -464.075398,
				       -451.365215,
				       -436.463959,
				       -420.438963,
				       -403.990574,
				       -387.584322,
				       -371.502351,
				       -355.923812,
				       -340.946128,
				       -326.615863,
				       -312.947150,
				       -299.932513,
				       -287.549580,
				       -275.769611,
				       -264.558115,
				       -253.877010,
				       -243.689490,
				       -233.956948,
				       -224.642118,
				       -215.708356,
				       -207.122175,
				       -198.836439,
				       -190.825146,
				       -183.045428,
				       -175.456039,
				       -168.007818,
				       -160.641735,
				       -153.280419,
				       -145.810654,
				       -138.048945,
				       -129.655317,
				       -119.951993,
				       -109.456520,
				       -102.169916,
				       -97.141669,
				       -93.133006,
				       -89.704576,
				       -86.667909,
				       -83.924058,
				       -81.423497,
				       -79.172024,
				       -77.099570,
				       -75.163161,
				       -73.342901,
				       -71.624602,
				       -69.997257,
				       -68.452046,
				       -66.981557,
				       -65.579433,
				       -64.240275,
				       -62.959259,
				       -61.732225,
				       -60.555390,
				       -59.425388,
				       -58.339212,
				       -57.294075,
				       -56.287474,
				       -55.317106,
				       -54.380864,
				       -53.476834,
				       -52.603212,
				       -51.758295,
				       -50.940582,
				       -50.148609,
				       -49.381060,
				       -48.636662,
				       -47.914216,
				       -47.212642,
				       -46.530897,
				       -45.867933,
				       -45.222887,
				       -44.594937,
				       -43.983259,
				       -43.387081,
				       -42.805668,
				       -42.238315,
				       -41.684341,
				       -41.143133,
				       -40.614095,
				       -40.096635,
				       -39.590201,
				       -39.094293,
				       -38.608399,
				       -38.132059
    };

  
    double Graph_Om_y_Im_T13 [150] = { 0.313878,
				       0.368610,
				       0.431055,
				       0.502141,
				       0.582931,
				       0.674837,
				       0.779583,
				       0.899176,
				       1.035959,
				       1.192666,
				       1.372513,
				       1.579273,
				       1.817399,
				       2.092184,
				       2.409804,
				       2.777770,
				       3.204856,
				       3.701555,
				       4.280654,
				       4.956692,
				       5.747987,
				       6.674972,
				       7.762763,
				       9.040322,
				       10.540098,
				       12.300707,
				       14.363690,
				       16.775194,
				       19.589250,
				       22.859508,
				       26.649545,
				       31.034828,
				       36.090930,
				       41.925515,
				       48.639537,
				       56.386859,
				       65.338516,
				       75.695877,
				       87.732655,
				       101.748067,
				       118.149329,
				       137.428874,
				       160.171127,
				       187.126538,
				       219.209043,
				       257.508320,
				       303.372286,
				       358.207936,
				       423.674585,
				       501.192578,
				       591.537257,
				       694.244016,
				       805.850223,
				       919.402082,
				       1023.20456,
				       1102.584176,
				       1144.255389,
				       1141.482241,
				       1097.291368,
				       1022.350076,
				       930.208991,
				       832.775961,
				       738.301212,
				       651.388388,
				       573.925390,
				       506.193188,
				       447.542576,
				       397.002511,
				       353.486102,
				       315.960002,
				       283.518367,
				       255.347314,
				       230.793268,
				       209.283499,
				       190.349776,
				       173.609728,
				       158.733608,
				       145.453946,
				       133.545571,
				       122.818811,
				       113.120402,
				       104.309579,
				       96.275395,
				       88.921457,
				       82.162287,
				       75.931660,
				       70.165414,
				       64.801224,
				       59.802191,
				       55.121033,
				       50.722625,
				       46.570594,
				       42.633929,
				       38.887847,
				       35.313340,
				       31.917492,
				       28.800893,
				       26.552234,
				       27.660677,
				       31.165483,
				       34.081518,
				       36.317887,
				       38.063159,
				       39.455256,
				       40.590090,
				       41.550560,
				       42.324927,
				       42.924916,
				       43.384620,
				       43.730864,
				       43.982760,
				       44.155337,
				       44.259936,
				       44.305865,
				       44.300907,
				       44.251282,
				       44.162455,
				       44.038929,
				       43.884433,
				       43.702406,
				       43.495663,
				       43.266757,
				       43.018004,
				       42.751156,
				       42.468056,
				       42.170154,
				       41.858889,
				       41.535493,
				       41.200979,
				       40.856391,
				       40.502626,
				       40.140464,
				       39.770688,
				       39.393967,
				       39.010874,
				       38.622006,
				       38.227975,
				       37.829254,
				       37.426253,
				       37.019384,
				       36.609005,
				       36.195493,
				       35.779189,
				       35.360378,
				       34.939372,
				       34.516464,
				       34.091907,
				       33.665937,
				       33.238826,
				       32.810787
    };

    TGraph *Graph_Rho_Re_T11 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Re_T11 );
    TGraph *Graph_Rho_Im_T11 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Im_T11 );
    TGraph *Graph_Rho_Re_T13 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Re_T13 );
    TGraph *Graph_Rho_Im_T13 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Im_T13 );
    TGraph *Graph_Rho_Re_T31 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Re_T31 );
    TGraph *Graph_Rho_Im_T31 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Im_T31 );
    TGraph *Graph_Rho_Re_T33 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Re_T33 );
    TGraph *Graph_Rho_Im_T33 = new TGraph(150, Graph_Rho_x, Graph_Rho_y_Im_T33 );

    TGraph *Graph_Om_Re_T11 = new TGraph(150, Graph_Om_x, Graph_Om_y_Re_T11 );
    TGraph *Graph_Om_Im_T11 = new TGraph(150, Graph_Om_x, Graph_Om_y_Im_T11 );
    TGraph *Graph_Om_Re_T13 = new TGraph(150, Graph_Om_x, Graph_Om_y_Re_T13 );
    TGraph *Graph_Om_Im_T13 = new TGraph(150, Graph_Om_x, Graph_Om_y_Im_T13 );
  
    Pi_minusBeamAmplitude->SetGraph_Rho_Re_T11(Graph_Rho_Re_T11);
    Pi_minusBeamAmplitude->SetGraph_Rho_Im_T11(Graph_Rho_Im_T11);
    Pi_minusBeamAmplitude->SetGraph_Rho_Re_T13(Graph_Rho_Re_T13);
    Pi_minusBeamAmplitude->SetGraph_Rho_Im_T13(Graph_Rho_Im_T13);
    Pi_minusBeamAmplitude->SetGraph_Rho_Re_T31(Graph_Rho_Re_T31);
    Pi_minusBeamAmplitude->SetGraph_Rho_Im_T31(Graph_Rho_Im_T31);
    Pi_minusBeamAmplitude->SetGraph_Rho_Re_T33(Graph_Rho_Re_T33);
    Pi_minusBeamAmplitude->SetGraph_Rho_Im_T33(Graph_Rho_Im_T33);
    Pi_minusBeamAmplitude->SetGraph_Om_Re_T11(Graph_Om_Re_T11);
    Pi_minusBeamAmplitude->SetGraph_Om_Im_T11(Graph_Om_Im_T11);
    Pi_minusBeamAmplitude->SetGraph_Om_Re_T13(Graph_Om_Re_T13);
    Pi_minusBeamAmplitude->SetGraph_Om_Im_T13(Graph_Om_Im_T13);

    Pi_minusBeamAmplitude->EnableWeighting();
    Pi_minusBeamAmplitude->SetExpectedWeightMean(-1);
    Pi_minusBeamAmplitude->SetVersionFlag(512);
  

    pdmutil->Add(Pi_minusBeamAmplitude);
  


    Pi_plusBeamAmplitude->SetGraph_Rho_Re_T11(Graph_Rho_Re_T11);
    Pi_plusBeamAmplitude->SetGraph_Rho_Im_T11(Graph_Rho_Im_T11);
    Pi_plusBeamAmplitude->SetGraph_Rho_Re_T13(Graph_Rho_Re_T13);
    Pi_plusBeamAmplitude->SetGraph_Rho_Im_T13(Graph_Rho_Im_T13);
    Pi_plusBeamAmplitude->SetGraph_Rho_Re_T31(Graph_Rho_Re_T31);
    Pi_plusBeamAmplitude->SetGraph_Rho_Im_T31(Graph_Rho_Im_T31);
    Pi_plusBeamAmplitude->SetGraph_Rho_Re_T33(Graph_Rho_Re_T33);
    Pi_plusBeamAmplitude->SetGraph_Rho_Im_T33(Graph_Rho_Im_T33);
    Pi_plusBeamAmplitude->SetGraph_Om_Re_T11(Graph_Om_Re_T11);
    Pi_plusBeamAmplitude->SetGraph_Om_Im_T11(Graph_Om_Im_T11);
    Pi_plusBeamAmplitude->SetGraph_Om_Re_T13(Graph_Om_Re_T13);
    Pi_plusBeamAmplitude->SetGraph_Om_Im_T13(Graph_Om_Im_T13);

    Pi_plusBeamAmplitude->EnableWeighting();
    Pi_plusBeamAmplitude->SetExpectedWeightMean(-1);
    Pi_plusBeamAmplitude->SetVersionFlag(512);

    pdmutil->Add(Pi_plusBeamAmplitude);

    if (Pi_plusBeamAmplitude && Pi_minusBeamAmplitude) return kTRUE;
    return kFALSE;
}

PPionBeamPlugin::~PPionBeamPlugin() {
}


Bool_t PPionBeamPlugin::ExecCommand(const char *, Double_t) {
    return kTRUE;
}



ClassImp(PPionBeamPlugin)



